# include("MPC/main.jl")

import Ipopt, JuMP, Plots

include("E. coli model/E_coli_Model.jl")
include("init.jl")
include("DA.jl")
include("pFBA reformulations/Duality.jl")
include("pFBA reformulations/P-Duality.jl")
include("pFBA reformulations/KKT.jl")
include("pFBA reformulations/P-KKT.jl")

# Initialize E. coli core metabolic network:   
mod = E_coli_Model_init
changeRxnBounds(mod, "EX_o2(e)", v_omax , "l")

# Initialize MPC:  
t_0 = 0   # h
t_end = 35 # h
N = 35 
dt = (t_end-t_0)/N
sp  = [0.2*ones(1,6) 0.4*ones(1,15) 0.2*ones(1,15)] # 0.2*ones(1,N+1)
G_f = G_f0*ones(1,N+1) # [G_f0*ones(1,6) 8*ones(1,15) G_f0*ones(1,15)] # G_f0*ones(1,N+1)
v_gmax = v_gmax_0*ones(1,N+1) #[v_gmax_0*ones(1,6) -5.25*ones(1,15) v_gmax_0*ones(1,15)]

time = Array{Float64}(undef, N+1)
time[1] = t_0

# Dual:
G_Dual =    Array{Float64}(undef, N+1)
X_Dual =    Array{Float64}(undef, N+1)
A_Dual =    Array{Float64}(undef, N+1)
D_Dual =    Array{Float64}(undef, N+1)
time_Dual = Array{Float64}(undef, N)
status_Dual = Array{Any}(undef, N)
t_list_Dual = [t_0]
G_list_Dual = [G_0]
X_list_Dual = [X_0]
A_list_Dual = [A_0]
G_Dual[1]    = G_0
X_Dual[1]    = X_0
A_Dual[1]    = A_0
D_Dual[1]    = D_0

# P. Dual:
G_PDual =    Array{Float64}(undef, N+1)
X_PDual =    Array{Float64}(undef, N+1)
A_PDual =    Array{Float64}(undef, N+1)
D_PDual =    Array{Float64}(undef, N+1)
time_PDual = Array{Float64}(undef, N)
status_PDual = Array{Any}(undef, N)
t_list_PDual = [t_0]
G_list_PDual = [G_0]
X_list_PDual = [X_0]
A_list_PDual = [A_0]
G_PDual[1]    = G_0
X_PDual[1]    = X_0
A_PDual[1]    = A_0
D_PDual[1]    = D_0

# KKT:
G_KKT =    Array{Float64}(undef, N+1)
X_KKT =    Array{Float64}(undef, N+1)
A_KKT =    Array{Float64}(undef, N+1)
D_KKT =    Array{Float64}(undef, N+1)
time_KKT = Array{Float64}(undef, N)
status_KKT = Array{Any}(undef, N)
t_list_KKT = [t_0]
G_list_KKT = [G_0]
X_list_KKT = [X_0]
A_list_KKT = [A_0]
G_KKT[1]    = G_0
X_KKT[1]    = X_0
A_KKT[1]    = A_0
D_KKT[1]    = D_0

# P. KKT:
G_PKKT =    Array{Float64}(undef, N+1)
X_PKKT =    Array{Float64}(undef, N+1)
A_PKKT =    Array{Float64}(undef, N+1)
D_PKKT =    Array{Float64}(undef, N+1)
time_PKKT = Array{Float64}(undef, N)
status_PKKT = Array{Any}(undef, N)
t_list_PKKT = [t_0]
G_list_PKKT = [G_0]
X_list_PKKT = [X_0]
A_list_PKKT = [A_0]
G_PKKT[1]    = G_0
X_PKKT[1]    = X_0
A_PKKT[1]    = A_0
D_PKKT[1]    = D_0

# MSE:
MSE_Dual = 0
MSE_PDual = 0
MSE_KKT = 0
MSE_PKKT = 0

# MPC: 
for n = 1:N
    global t_list_Dual, G_list_Dual, X_list_Dual, A_list_Dual, t_list_PDual, G_list_PDual, X_list_PDual, A_list_PDual, t_list_PKKT, G_list_PKKT, X_list_PKKT, A_list_PKKT, MSE_Dual, MSE_PDual, MSE_PKKT, time_Dual, time_PDual, time_PKKT,status_Dual, status_PDual, status_PKKT, status_KKT, t_list_KKT, G_list_KKT, X_list_KKT, A_list_KKT, MSE_KKT
    display(n)

    time[n+1] = time[n]+dt

    # MPC control optimizations:
    D_Dual[n+1], time_Dual[n], status_Dual[n] = Dual(mod, G_Dual[n], X_Dual[n], A_Dual[n], K_g, v_gmax[n], 4, D_Dual[n], G_f[n], sp[n])
    D_PDual[n+1], time_PDual[n], status_PDual[n] = PDual(mod, G_PDual[n], X_PDual[n], A_PDual[n], K_g, v_gmax[n], 4, D_PDual[n], G_f[n], sp[n]) 
    D_KKT[n+1], time_KKT[n], status_KKT[n] = KKT(mod, G_KKT[n], X_KKT[n], A_KKT[n], K_g, v_gmax[n], 4, D_KKT[n], G_f[n], sp[n])
    D_PKKT[n+1], time_PKKT[n], status_PKKT[n] = PKKT(mod, G_PKKT[n], X_PKKT[n], A_PKKT[n], K_g, v_gmax[n], 4, D_PKKT[n], G_f[n], sp[n])

    # Dual:
    G_Dual[n+1], X_Dual[n+1], A_Dual[n+1], C_t, C_G, C_X, C_A = DAP(mod, G_Dual[n], X_Dual[n], A_Dual[n], K_g, v_gmax[n], v_omax, time[n], time[n+1], D_Dual[n+1], G_f[n])
    t_list_Dual = vcat(t_list_Dual, C_t)
    G_list_Dual = vcat(G_list_Dual, C_G)
    X_list_Dual = vcat(X_list_Dual, C_X)
    A_list_Dual = vcat(A_list_Dual, C_A)
    
    # P. Dual:
    G_PDual[n+1], X_PDual[n+1], A_PDual[n+1], C_t, C_G, C_X, C_A = DAP(mod, G_PDual[n], X_PDual[n], A_PDual[n], K_g, v_gmax[n], v_omax, time[n], time[n+1], D_PDual[n+1], G_f[n])
    t_list_PDual = vcat(t_list_PDual, C_t)
    G_list_PDual = vcat(G_list_PDual, C_G)
    X_list_PDual = vcat(X_list_PDual, C_X)
    A_list_PDual = vcat(A_list_PDual, C_A)
    
    # KKT:
    G_KKT[n+1], X_KKT[n+1], A_KKT[n+1], C_t, C_G, C_X, C_A = DAP(mod, G_KKT[n], X_KKT[n], A_KKT[n], K_g, v_gmax[n], v_omax, time[n], time[n+1], D_KKT[n+1], G_f[n])
    t_list_KKT = vcat(t_list_KKT, C_t)
    G_list_KKT = vcat(G_list_KKT, C_G)
    X_list_KKT = vcat(X_list_KKT, C_X)
    A_list_KKT = vcat(A_list_KKT, C_A)
    
    # P. KKT:
    G_PKKT[n+1], X_PKKT[n+1], A_PKKT[n+1], C_t, C_G, C_X, C_A = DAP(mod, G_PKKT[n], X_PKKT[n], A_PKKT[n], K_g, v_gmax[n], v_omax, time[n], time[n+1], D_PKKT[n+1], G_f[n])
    t_list_PKKT = vcat(t_list_PKKT, C_t)
    G_list_PKKT = vcat(G_list_PKKT, C_G)
    X_list_PKKT = vcat(X_list_PKKT, C_X)
    A_list_PKKT = vcat(A_list_PKKT, C_A)

    # MSE:
    MSE_Dual = MSE_Dual + (X_Dual[n]- sp[n])^2
    MSE_PDual = MSE_PDual + (X_PDual[n]- sp[n])^2
    MSE_KKT = MSE_KKT + (X_KKT[n]- sp[n])^2
    MSE_PKKT = MSE_PKKT + (X_PKKT[n]- sp[n])^2
end

# Calculate MSE:
MSE_Dual = MSE_Dual + (X_Dual[N+1]- sp[N+1])^2
MSE_PDual = MSE_PDual + (X_PDual[N+1]- sp[N+1])^2
MSE_KKT = MSE_KKT + (X_KKT[N+1]- sp[N+1])^2
MSE_PKKT = MSE_PKKT + (X_PKKT[N+1]- sp[N+1])^2

# Display results: 
display("Status:")
display("Dual:")
for i in status_Dual
    if i != "Solve_Succeeded"
        display(i)
    end
end
display("P. Dual:")
for i in status_PDual
    if i != "Solve_Succeeded"
        display(i)
    end
end
display("KKT:")
for i in status_KKT
    if i != "Solve_Succeeded"
        display(i)
    end
end
display("P. KKT:")
for i in status_PKKT
    if i != "Solve_Succeeded"
        display(i)
    end
end

display("MSE:")
display(MSE_Dual/N)
display(MSE_PDual/N)
display(MSE_KKT/N)
display(MSE_PKKT/N)

display("Time:")
display(sum(time_Dual)/N)
display(sum(time_PDual)/N)
display(sum(time_KKT)/N)
display(sum(time_PKKT)/N)

# Plot results: 
using Plots

pG =  plot(t_list_Dual,G_list_Dual, title="Glucose", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="Dual", lw=2, ls=:solid, color=:orange)
pG =  plot!(t_list_PDual,G_list_PDual, title="Glucose", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="P. Dual", lw=2, ls=:dash, color=:green)
pG =  plot!(t_list_KKT,G_list_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pG =  plot!(t_list_PKKT,G_list_PKKT, label="P. KKT", lw=2, ls=:dot, color=:red)

pX =  plot(t_list_Dual,X_list_Dual, title="Biomass", xaxis="Time [h]",yaxis="Concentration [g/L]", label="Dual", lw=2, ls=:solid, color=:orange)
pX =  plot!(t_list_PDual,X_list_PDual, title="Biomass", xaxis="Time [h]",yaxis="Concentration [g/L]", label="P. Dual", lw=2, ls=:dash, color=:green)
pX = plot!(t_list_KKT,X_list_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pX =  plot!(t_list_PKKT,X_list_PKKT, label="P. KKT", lw=2, ls=:dot, color=:red)
pX =  plot!(time, sp[1,:], label="Setpoint", lw=2, ls=:dash, color=:blue, linetype=:steppost)

pA =  plot(t_list_Dual,A_list_Dual, title="Acetate", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="Dual", lw=2, ls=:solid, color=:orange)
pA =  plot!(t_list_PDual,A_list_PDual, title="Acetate", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="P. Dual", lw=2, ls=:dash, color=:green)
pA = plot!(t_list_KKT,A_list_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pA =  plot!(t_list_PKKT,A_list_PKKT, label="P. KKT", lw=2, ls=:dot, color=:red)

pD =  plot(time, D_Dual, title="Dilution rate", xaxis="Time [h]",yaxis="Dilution rate [h^-1]", label="Dual", lw=2, ls=:solid, color=:orange, linetype=:steppost)
pD =  plot!(time, D_PDual, title="Dilution rate", xaxis="Time [h]",yaxis="Dilution rate [h^-1]", label="P. Dual", lw=2, ls=:dash, color=:green, linetype=:steppost)
pD =  plot!(time, D_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple, linetype=:steppost)
pD =  plot!(time, D_PKKT, label="P. KKT", lw=2, ls=:dot, color=:red, linetype=:steppost)

pGf =  plot(time, G_f[1,:], title="Glucose feed", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="Glucose feed", lw=2, ls=:dash, color=:blue, linetype=:steppost)

pgmax =  plot(time, v_gmax[1,:], title="Maximal glucose uptake", xaxis="Time [h]",yaxis="Maximal glucose uptake", label="Maximal glucose uptake", lw=2, ls=:dash, color=:blue, linetype=:steppost)

# Save plots: 
savefig(pG,"MPC-G.pdf")
savefig(pX,"MPC-X.pdf")
savefig(pA,"MPC-A.pdf")
savefig(pD,"MPC-D.pdf")
savefig(pGf,"MPC-Gf.pdf")
savefig(pgmax,"MPC-gmax.pdf")