# include("dFBA Batch/main adaptive mesh.jl")

import Ipopt, JuMP, Plots, TimerOutputs, DifferentialEquations

include("E. coli core/E_coli_Model.jl")
include("init.jl")
include("DAP/DAP.jl")
include("DAP/DAPsaveat.jl")
include("MSE.jl")

include("Adaptive mesh/Duality.jl")
include("Adaptive mesh/P-Duality.jl")
include("Adaptive mesh/KKT.jl")
include("Adaptive mesh/P-KKT.jl")

timer = TimerOutputs.TimerOutput()

# Initialize model:
mod = E_coli_model
changeRxnBounds(mod, "EX_o2(e)", v_omax , "l")

Bio_index =  13
Ace_index =  20
Glu_index =  28

t_0 = 0     # h
t_end = 7.1 # h
N = 10      # Number of finite elements

# Dual:
display("Dual")
TimerOutputs.reset_timer!(timer)
t_Dual, G_Dual, X_Dual, A_Dual, status_Dual = TDual(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_0, t_end)
display(timer)

# P. Dual:
display("P. Dual:")
TimerOutputs.reset_timer!(timer)
t_PDual, G_PDual, X_PDual, A_PDual, status_PDual = TPDual(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_0, t_end)
display(timer)

# KKT:
display("KKT")
TimerOutputs.reset_timer!(timer)
t_KKT, G_KKT, X_KKT, A_KKT, status_KKT = TKKT(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_0, t_end)
display(timer)

# P. KKT:
display("P. KKT")
TimerOutputs.reset_timer!(timer)
t_PKKT, G_PKKT, X_PKKT, A_PKKT, status_PKKT = TPKKT(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_0, t_end)
display(timer)

# Display results:

t_DAP, G_DAP, X_DAP, A_DAP = DAPsave(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end, t_Dual)
display("Dual:")
display(status_Dual)
display(MSE(G_DAP, G_Dual))
display(MSE(X_DAP, X_Dual))
display(MSE(A_DAP, A_Dual))

t_DAP, G_DAP, X_DAP, A_DAP = DAPsave(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end, t_PDual)
display("P. Dual:")
display(status_PDual)
display(MSE(G_DAP, G_PDual))
display(MSE(X_DAP, X_PDual))
display(MSE(A_DAP, A_PDual))

t_DAP, G_DAP, X_DAP, A_DAP = DAPsave(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end, t_KKT)
display("KKT:")
display(status_KKT)
display(MSE(G_DAP, G_KKT))
display(MSE(X_DAP, X_KKT))
display(MSE(A_DAP, A_KKT))

t_DAP, G_DAP, X_DAP, A_DAP = DAPsave(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end, t_PKKT)
display("P. KKT:")
display(status_PKKT)
display(MSE(G_DAP, G_PKKT))
display(MSE(X_DAP, X_PKKT))
display(MSE(A_DAP, A_PKKT))

# Direct Approach:
display("DA:")
TimerOutputs.reset_timer!(timer)
t_DAP, G_DAP, X_DAP, A_DAP = DAP(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end)
display(timer)

# Ploting the results:
using Plots

t_points_Dual = [t_0]
for n = 4:3:N*3+1
    global t_points_Dual
    t_points_Dual = vcat(t_points_Dual, t_Dual[n])
end

t_points = [t_0]
for n = 4:3:N*3+1
    global t_points
    t_points = vcat(t_points, t_PDual[n])
end

t_points_KKT = [t_0]
for n = 4:3:N*3+1
    global t_points_KKT
    t_points_KKT = vcat(t_points_KKT, t_KKT[n])
end

t_points_PKKT = [t_0]
for n = 4:3:N*3+1
    global t_points_PKKT
    t_points_PKKT = vcat(t_points_PKKT, t_PKKT[n])
end

pG = plot(t_DAP,G_DAP, title="Glucose", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="DA", lw=2, ls=:solid, color=:blue)
pG = plot!(t_Dual,G_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pG = plot!(t_PDual,G_PDual, label="P. Dual", lw=2, ls=:dash, color=:green)
pG = plot!(t_KKT,G_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pG = plot!(t_PKKT,G_PKKT, label="P. KKT", lw=2, ls=:dot, color=:red)
pG = scatter!(t_points_Dual, zeros(length(t_points_Dual)) ,label="Finite elements Dual",color=:orange, markershape=:star5)
pG = scatter!(t_points, zeros(length(t_points)) , label="Finite elements P. Dual", color=:green)
pG = scatter!(t_points_KKT, zeros(length(t_points_KKT)) ,label="Finite elements KKT",color=:purple, markershape=:square)
pG = scatter!(t_points_PKKT, zeros(length(t_points_PKKT)) ,label="Finite elements P. KKT",color=:red, markershape=:cross)

pX = plot(t_DAP,X_DAP, title="Biomass", xaxis="Time [h]",yaxis="Concentration [g/L]", label="DA", lw=2, ls=:solid, color=:blue)
pX = plot!(t_Dual,X_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pX = plot!(t_PDual,X_PDual, label="P. Dual", lw=2, ls=:dash, color=:green)
pX = plot!(t_KKT,X_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pX = plot!(t_PKKT,X_PKKT, label="P. KKT", lw=2, ls=:dot, color=:red)
pX = scatter!(t_points_Dual, zeros(length(t_points_Dual)) ,label="Finite elements Dual",color=:orange, markershape=:star5)
pX = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements P. Dual", color=:green)
pX = scatter!(t_points_KKT, zeros(length(t_points_KKT)) ,label="Finite elements KKT",color=:purple, markershape=:square)
pX = scatter!(t_points_PKKT, zeros(length(t_points_PKKT)) ,label="Finite elements P. KKT",color=:red, markershape=:cross)

pA = plot(t_DAP,A_DAP, title="Acetate", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="DA", lw=2, ls=:solid, color=:blue, legend=:topleft)
pA = plot!(t_Dual,A_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pA = plot!(t_PDual,A_PDual, label="P. Dual", lw=2, ls=:dash, color=:green)
pA = plot!(t_KKT,A_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pA = plot!(t_PKKT,A_PKKT, label="P. KKT", lw=2, ls=:dot, color=:red)
pA = scatter!(t_points_Dual, zeros(length(t_points_Dual)) ,label="Finite elements Dual",color=:orange, markershape=:star5)
pA = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements P. Dual", color=:green)
pA = scatter!(t_points_KKT, zeros(length(t_points_KKT)) ,label="Finite elements KKT",color=:purple, markershape=:square)
pA = scatter!(t_points_PKKT, zeros(length(t_points_PKKT)) ,label="Finite elements P. KKT",color=:red, markershape=:cross)

# Save plots:
savefig(pG,"SGT.pdf")
savefig(pX,"SXT.pdf")
savefig(pA,"SAT.pdf")
