# include("dFBA Batch/main flux.jl")

import Ipopt, JuMP, Plots, TimerOutputs, DifferentialEquations

include("E. coli core/E_coli_Model.jl")
include("init.jl")
include("DAP/DAP.jl")
include("DAP/DAPsaveat.jl")
include("MSE.jl")

include("Flux constraints/Duality.jl")
include("Flux constraints/PDuality.jl")
include("Flux constraints/PKKT.jl")

timer = TimerOutputs.TimerOutput()

# Initialize model:
mod = E_coli_model
changeRxnBounds(mod, "EX_o2(e)", v_omax , "l")

Bio_index =  13
Ace_index =  20
Glu_index =  28

t_0 = 0     # h
t_end = 7.1 # h

N = 10              # Number of finite elements    
dt = (t_end-t_0)/N  # Size of finite elements

# Optimization problems

display("SDual")
TimerOutputs.reset_timer!(timer)
t_Dual, G_Dual, X_Dual, A_Dual, status_Dual = SDual(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, dt)
display(timer)

display("SPDual")
TimerOutputs.reset_timer!(timer)
t_PDual, G_PDual, X_PDual, A_PDual, status_PDual = SPDual(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, dt)
display(timer)

display("SPKKT")
TimerOutputs.reset_timer!(timer)
t_PKKT, G_PKKT, X_PKKT, A_PKKT, status_PKKT = SPKKT(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, dt)
display(timer)

# Direct approach:

display("DAP")
TimerOutputs.reset_timer!(timer)
t_DAP, G_DAP, X_DAP, A_DAP = DAPsave(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end, t_PKKT)
display(timer)

# Display results:

display("SDual:")
display(status_Dual)
display(MSE(G_DAP, G_Dual))
display(MSE(X_DAP, X_Dual))
display(MSE(A_DAP, A_Dual))

display("SPDual:")
display(status_PDual)
display(MSE(G_DAP, G_PDual))
display(MSE(X_DAP, X_PDual))
display(MSE(A_DAP, A_PDual))

display("SPKKT:")
display(status_PKKT)
display(MSE(G_DAP, G_PKKT))
display(MSE(X_DAP, X_PKKT))
display(MSE(A_DAP, A_PKKT))


# Plot results:

display("DAP")
TimerOutputs.reset_timer!(timer)
t_DAP, G_DAP, X_DAP, A_DAP = DAP(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end)
display(timer)

using Plots
t_points = [t_PDual[1]]
for n = 4:3:N*3+1
    global t_points
    t_points = vcat(t_points, t_PDual[n])
end

pG = plot(t_DAP,G_DAP, title="Glucose", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="DA", lw=2, ls=:solid, color=:blue)
pG = plot!(t_Dual,G_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pG = plot!([t_PDual t_PKKT],[G_PDual G_PKKT], label=["P. Dual" "P. KKT"], lw=2, ls=[:dash :dot], color=[:green :red])
pG = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

pX = plot(t_DAP,X_DAP, title="Biomass", xaxis="Time [h]",yaxis="Concentration [g/L]", label="DA", lw=2, ls=:solid, color=:blue)
pX = plot!(t_Dual,X_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pX = plot!([t_PDual t_PKKT],[X_PDual X_PKKT], label=["P. Dual" "P. KKT"], lw=2, ls=[:dash :dot], color=[:green :red])
pX = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

pA = plot(t_DAP,A_DAP, title="Acetate", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="DA", lw=2, ls=:solid, color=:blue)
pA = plot!(t_Dual,A_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pA = plot!([t_PDual t_PKKT],[A_PDual A_PKKT], label=["P. Dual" "P. KKT"], lw=2, ls=[:dash :dot], color=[:green :red])
pA = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

savefig(pG,"SG.pdf")
savefig(pX,"SX.pdf")
savefig(pA,"SA.pdf")
