# include("dFBA Batch/main.jl")

import Ipopt, JuMP, Plots, TimerOutputs, DifferentialEquations

include("E. coli core/E_coli_Model.jl")
include("init.jl")
include("DAP/DAP.jl")
include("DAP/DAPsaveat.jl")
include("MSE.jl")

include("Test - Point/PDuality1.jl")
include("Test - Point/PDuality2.jl")
include("Test - Point/PDuality3.jl")

timer = TimerOutputs.TimerOutput()

# Initialize model:
mod = E_coli_Model_init # E. coli model - core metabolism
changeRxnBounds(mod, "EX_o2(e)", v_omax , "l")

Bio_index =  13
Ace_index =  20
Glu_index =  28

t_0 = 0   # h
t_end = 7.1 # h

N = 10              # Number of horizion point     
dt = (t_end-t_0)/N  # 

# Single optimization problems

display("SPDual1")
TimerOutputs.reset_timer!(timer)
t_PDual1, G_PDual1, X_PDual1, A_PDual1, status_PDual1, v, LB = SPDual1(mod, G_0, X_0, A_0, K_g, v_gmax)
display(timer)

display("SPDual2")
TimerOutputs.reset_timer!(timer)
t_PDual2, G_PDual2, X_PDual2, A_PDual2, status_PDual2, v, LB = SPDual2(mod, G_0, X_0, A_0, K_g, v_gmax)
display(timer)

display("SPDual3")
TimerOutputs.reset_timer!(timer)
t_PDual3, G_PDual3, X_PDual3, A_PDual3, status_PDual3, v, LB = SPDual3(mod, G_0, X_0, A_0, K_g, v_gmax)
display(timer)

# Direct approach:
display("DAP")
TimerOutputs.reset_timer!(timer)
t_DAP, G_DAP, X_DAP, A_DAP = DAPsave(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end, t_PDual1)
changeRxnBounds(mod, "EX_o2(e)", v_omax , "l")
display(timer)

# Display results:

display("SPDual1:")
display(status_PDual1)
display(MSE(G_DAP, G_PDual1))
display(MSE(X_DAP, X_PDual1))
display(MSE(A_DAP, A_PDual1))

display("SPDual2:")
display(status_PDual2)
display(MSE(G_DAP, G_PDual2))
display(MSE(X_DAP, X_PDual2))
display(MSE(A_DAP, A_PDual2))

display("SPDual3:")
display(status_PDual3)
display(MSE(G_DAP, G_PDual3))
display(MSE(X_DAP, X_PDual3))
display(MSE(A_DAP, A_PDual3))

display("DAP")
TimerOutputs.reset_timer!(timer)
t_DAP, G_DAP, X_DAP, A_DAP = DAP(mod, G_0, X_0, A_0, K_g, v_gmax, v_omax, t_end)
display(timer)

# Plot results:
using Plots
t_points = [t_PDual[1]]
for n = 4:3:N*3+1
    global t_points
    t_points = vcat(t_points, t_PDual[n])
end

pG =  plot(t_DAP,G_DAP, title="Glucose", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="DAP", lw=2, ls=:solid, color=:blue)
pG = plot!(t_PDual1,G_PDual1, label="P. Dual 1", lw=2, ls=:dashdot, color=:orange)
pG = plot!([t_PDual2 t_PDual3],[G_PDual2 G_PDual3], label=["P. Dual 2" "P. Dual 3"], lw=2, ls=[:dash :dot], color=[:green :red])
pG = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

pX =  plot(t_DAP,X_DAP, title="Biomass", xaxis="Time [h]",yaxis="Concentration [g/L]", label="DAP", lw=2, ls=:solid, color=:blue)
pX = plot!(t_PDual1,X_PDual1, label="P. Dual 1", lw=2, ls=:dashdot, color=:orange)
pX = plot!([t_PDual2 t_PDual3],[X_PDual2 X_PDual3], label=["P. Dual 2" "P. Dual 3"], lw=2, ls=[:dash :dot], color=[:green :red])
pX = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

pA =  plot(t_DAP,A_DAP, title="Acetate", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="DAP", lw=2, ls=:solid, color=:blue)
pA = plot!(t_PDual1,A_PDual1, label="P. Dual 1", lw=2, ls=:dashdot, color=:orange)
pA = plot!([t_PDual2 t_PDual3],[A_PDual2 A_PDual3], label=["P. Dual 2" "P. Dual 3"], lw=2, ls=[:dash :dot], color=[:green :red])
pA = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

savefig(pG,"SG123.pdf")
savefig(pX,"SX123.pdf")
savefig(pA,"SA123.pdf")
