# include("dFBA CSTR/main fixed.jl")

import Ipopt, JuMP, Plots, TimerOutputs, DifferentialEquations

include("E. coli core/E_coli_Model.jl")
include("init.jl")
include("DAP/DAP.jl")
include("DAP/DAPsaveat.jl")
include("MSE.jl")

include("Fixed time steps/Duality.jl")
include("Fixed time steps/PDuality.jl")
include("Fixed time steps/KKT.jl")
include("Fixed time steps/PKKT.jl")

timer = TimerOutputs.TimerOutput()

# Initialize model:
mod = E_coli_model # E. coli model - core metabolism
changeRxnBounds(mod, "EX_o2(e)", v_omax , "l")

t_0 = 0   # h
t_end = 30 # h
Bio_index =  13
Ace_index =  20
Glu_index =  28

N = 20
dt = (t_end-t_0)/N

# Single optimization problems

display("SDual")
TimerOutputs.reset_timer!(timer)
t_Dual, G_Dual, X_Dual, A_Dual, status_Dual = SDual(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, dt, G_f)
display(timer)

display("SPDual")
TimerOutputs.reset_timer!(timer)
t_PDual, G_PDual, X_PDual, A_PDual, status_PDual = SPDual(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, dt, G_f)
display(timer)

display("KKT")
TimerOutputs.reset_timer!(timer)
t_KKT, G_KKT, X_KKT, A_KKT, status_KKT = KKT(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, dt, G_f)
display(timer)

display("SPKKT")
TimerOutputs.reset_timer!(timer)
t_PKKT, G_PKKT, X_PKKT, A_PKKT, status_PKKT = SPKKT(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, dt, G_f)
display(timer)

# Direct approach:

display("DAP")
TimerOutputs.reset_timer!(timer)
t_DAP, G_DAP, X_DAP, A_DAP = DAPsave(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end, t_PKKT)
display(timer)

# Display results:

display("SDual:")
display(status_Dual)
display(MSE(G_DAP, G_Dual))
display(MSE(X_DAP, X_Dual))
display(MSE(A_DAP, A_Dual))

display("SPDual:")
display(status_PDual)
display(MSE(G_DAP, G_PDual))
display(MSE(X_DAP, X_PDual))
display(MSE(A_DAP, A_PDual))

display("KKT:")
display(status_KKT)
display(MSE(G_DAP, G_KKT))
display(MSE(X_DAP, X_KKT))
display(MSE(A_DAP, A_KKT))

display("SPKKT:")
display(status_PKKT)
display(MSE(G_DAP, G_PKKT))
display(MSE(X_DAP, X_PKKT))
display(MSE(A_DAP, A_PKKT))

display("DAP")
TimerOutputs.reset_timer!(timer)
t_DAP, G_DAP, X_DAP, A_DAP = DAP(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end)
display(timer)

# Plot results:
using Plots

t_points = [t_0]
for n = 4:3:N*3+1
    global t_points
    t_points = vcat(t_points, t_PDual[n])
end

pG = plot(t_DAP,G_DAP, title="Glucose", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="DA", lw=2, ls=:solid, color=:blue)
pG = plot!(t_Dual,G_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pG = plot!(t_Dual,G_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pG = plot!([t_PDual t_PKKT],[G_PDual G_PKKT], label=["P. Dual" "P. KKT"], lw=2, ls=[:dash :dot], color=[:green :red])
pG = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

pX = plot(t_DAP,X_DAP, title="Biomass", xaxis="Time [h]",yaxis="Concentration [g/L]", label="DA", lw=2, ls=:solid, color=:blue)
pX = plot!(t_Dual,X_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pX = plot!(t_Dual,X_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pX = plot!([t_PDual t_PKKT],[X_PDual X_PKKT], label=["P. Dual" "P. KKT"], lw=2, ls=[:dash :dot], color=[:green :red])
pX = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

pA = plot(t_DAP,A_DAP, title="Acetate", xaxis="Time [h]",yaxis="Concentration [mmol/L]", label="DA", lw=2, ls=:solid, color=:blue)
pA = plot!(t_Dual,A_Dual, label="Dual", lw=2, ls=:dashdot, color=:orange)
pA = plot!(t_Dual,A_KKT, label="KKT", lw=2, ls=:dashdotdot, color=:purple)
pA = plot!([t_PDual t_PKKT],[A_PDual A_PKKT], label=["P. Dual" "P. KKT"], lw=2, ls=[:dash :dot], color=[:green :red])
pA = scatter!(t_points, zeros(length(t_points)) ,label="Finite elements",color=:black)

savefig(pG,"SG.pdf")
savefig(pX,"SX.pdf")
savefig(pA,"SA.pdf")

display(last(G_DAP))
display(last(X_DAP))
display(last(A_DAP))